@@ -2,6 +2,7 @@ package qlog
 
 import (
 	"bytes"
+	"context"
 	"encoding/json"
 	"errors"
 	"io"
@@ -10,9 +11,10 @@ import (
 	"os"
 	"time"
 
-	"github.com/lucas-clemente/quic-go/internal/utils"
-
+	"github.com/lucas-clemente/quic-go"
 	"github.com/lucas-clemente/quic-go/internal/protocol"
+	"github.com/lucas-clemente/quic-go/internal/qerr"
+	"github.com/lucas-clemente/quic-go/internal/utils"
 	"github.com/lucas-clemente/quic-go/logging"
 
 	. "github.com/onsi/ginkgo"
@@ -43,20 +45,37 @@ func (w *limitedWriter) Write(p []byte) (int, error) {
 }
 
 type entry struct {
-	Time     time.Time
-	Category string
-	Name     string
-	Event    map[string]interface{}
+	Time  time.Time
+	Name  string
+	Event map[string]interface{}
 }
 
 var _ = Describe("Tracing", func() {
 	Context("tracer", func() {
 		It("returns nil when there's no io.WriteCloser", func() {
 			t := NewTracer(func(logging.Perspective, []byte) io.WriteCloser { return nil })
-			Expect(t.TracerForConnection(logging.PerspectiveClient, logging.ConnectionID{1, 2, 3, 4})).To(BeNil())
+			Expect(t.TracerForConnection(context.Background(), logging.PerspectiveClient, logging.ConnectionID{1, 2, 3, 4})).To(BeNil())
 		})
 	})
 
+	It("stops writing when encountering an error", func() {
+		buf := &bytes.Buffer{}
+		t := NewConnectionTracer(
+			&limitedWriter{WriteCloser: nopWriteCloser(buf), N: 250},
+			protocol.PerspectiveServer,
+			protocol.ConnectionID{0xde, 0xad, 0xbe, 0xef},
+		)
+		for i := uint32(0); i < 1000; i++ {
+			t.UpdatedPTOCount(i)
+		}
+
+		b := &bytes.Buffer{}
+		log.SetOutput(b)
+		defer log.SetOutput(os.Stdout)
+		t.Close()
+		Expect(b.String()).To(ContainSubstring("writer full"))
+	})
+
 	Context("connection tracer", func() {
 		var (
 			tracer logging.ConnectionTracer
@@ -66,7 +85,7 @@ var _ = Describe("Tracing", func() {
 		BeforeEach(func() {
 			buf = &bytes.Buffer{}
 			t := NewTracer(func(logging.Perspective, []byte) io.WriteCloser { return nopWriteCloser(buf) })
-			tracer = t.TracerForConnection(logging.PerspectiveServer, logging.ConnectionID{0xde, 0xad, 0xbe, 0xef})
+			tracer = t.TracerForConnection(context.Background(), logging.PerspectiveServer, logging.ConnectionID{0xde, 0xad, 0xbe, 0xef})
 		})
 
 		It("exports a trace that has the right metadata", func() {
@@ -74,12 +93,10 @@ var _ = Describe("Tracing", func() {
 
 			m := make(map[string]interface{})
 			Expect(json.Unmarshal(buf.Bytes(), &m)).To(Succeed())
-			Expect(m).To(HaveKeyWithValue("qlog_version", "draft-02-wip"))
+			Expect(m).To(HaveKeyWithValue("qlog_version", "draft-02"))
 			Expect(m).To(HaveKey("title"))
-			Expect(m).To(HaveKey("traces"))
-			traces := m["traces"].([]interface{})
-			Expect(traces).To(HaveLen(1))
-			trace := traces[0].(map[string]interface{})
+			Expect(m).To(HaveKey("trace"))
+			trace := m["trace"].(map[string]interface{})
 			Expect(trace).To(HaveKey(("common_fields")))
 			commonFields := trace["common_fields"].(map[string]interface{})
 			Expect(commonFields).To(HaveKeyWithValue("ODCID", "deadbeef"))
@@ -87,56 +104,42 @@ var _ = Describe("Tracing", func() {
 			Expect(commonFields).To(HaveKey("reference_time"))
 			referenceTime := time.Unix(0, int64(commonFields["reference_time"].(float64)*1e6))
 			Expect(referenceTime).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-			Expect(trace).To(HaveKey("event_fields"))
-			for i, ef := range trace["event_fields"].([]interface{}) {
-				Expect(ef.(string)).To(Equal(eventFields[i]))
-			}
+			Expect(commonFields).To(HaveKeyWithValue("time_format", "relative"))
 			Expect(trace).To(HaveKey("vantage_point"))
 			vantagePoint := trace["vantage_point"].(map[string]interface{})
 			Expect(vantagePoint).To(HaveKeyWithValue("type", "server"))
 		})
 
-		It("stops writing when encountering an error", func() {
-			tracer = newConnectionTracer(
-				&limitedWriter{WriteCloser: nopWriteCloser(buf), N: 250},
-				protocol.PerspectiveServer,
-				protocol.ConnectionID{0xde, 0xad, 0xbe, 0xef},
-			)
-			for i := uint32(0); i < 1000; i++ {
-				tracer.UpdatedPTOCount(i)
-			}
-
-			buf := &bytes.Buffer{}
-			log.SetOutput(buf)
-			defer log.SetOutput(os.Stdout)
-			tracer.Close()
-			Expect(buf.String()).To(ContainSubstring("writer full"))
-		})
-
 		Context("Events", func() {
 			exportAndParse := func() []entry {
 				tracer.Close()
 
 				m := make(map[string]interface{})
-				Expect(json.Unmarshal(buf.Bytes(), &m)).To(Succeed())
-				Expect(m).To(HaveKey("traces"))
+				line, err := buf.ReadBytes('\n')
+				Expect(err).ToNot(HaveOccurred())
+				Expect(json.Unmarshal(line, &m)).To(Succeed())
+				Expect(m).To(HaveKey("trace"))
 				var entries []entry
-				traces := m["traces"].([]interface{})
-				Expect(traces).To(HaveLen(1))
-				trace := traces[0].(map[string]interface{})
+				trace := m["trace"].(map[string]interface{})
 				Expect(trace).To(HaveKey("common_fields"))
 				commonFields := trace["common_fields"].(map[string]interface{})
 				Expect(commonFields).To(HaveKey("reference_time"))
 				referenceTime := time.Unix(0, int64(commonFields["reference_time"].(float64)*1e6))
-				Expect(trace).To(HaveKey("events"))
-				for _, e := range trace["events"].([]interface{}) {
-					ev := e.([]interface{})
-					Expect(ev).To(HaveLen(4))
+				Expect(trace).ToNot(HaveKey("events"))
+
+				for buf.Len() > 0 {
+					line, err := buf.ReadBytes('\n')
+					Expect(err).ToNot(HaveOccurred())
+					ev := make(map[string]interface{})
+					Expect(json.Unmarshal(line, &ev)).To(Succeed())
+					Expect(ev).To(HaveLen(3))
+					Expect(ev).To(HaveKey("time"))
+					Expect(ev).To(HaveKey("name"))
+					Expect(ev).To(HaveKey("data"))
 					entries = append(entries, entry{
-						Time:     referenceTime.Add(time.Duration(ev[0].(float64)*1e6) * time.Nanosecond),
-						Category: ev[1].(string),
-						Name:     ev[2].(string),
-						Event:    ev[3].(map[string]interface{}),
+						Time:  referenceTime.Add(time.Duration(ev["time"].(float64)*1e6) * time.Nanosecond),
+						Name:  ev["name"].(string),
+						Event: ev["data"].(map[string]interface{}),
 					})
 				}
 				return entries
@@ -152,47 +155,124 @@ var _ = Describe("Tracing", func() {
 				tracer.StartedConnection(
 					&net.UDPAddr{IP: net.IPv4(192, 168, 13, 37), Port: 42},
 					&net.UDPAddr{IP: net.IPv4(192, 168, 12, 34), Port: 24},
-					0xdeadbeef,
 					protocol.ConnectionID{1, 2, 3, 4},
 					protocol.ConnectionID{5, 6, 7, 8},
 				)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("connection_started"))
+				Expect(entry.Name).To(Equal("transport:connection_started"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("ip_version", "ipv4"))
 				Expect(ev).To(HaveKeyWithValue("src_ip", "192.168.13.37"))
 				Expect(ev).To(HaveKeyWithValue("src_port", float64(42)))
 				Expect(ev).To(HaveKeyWithValue("dst_ip", "192.168.12.34"))
 				Expect(ev).To(HaveKeyWithValue("dst_port", float64(24)))
-				Expect(ev).To(HaveKeyWithValue("quic_version", "deadbeef"))
 				Expect(ev).To(HaveKeyWithValue("src_cid", "01020304"))
 				Expect(ev).To(HaveKeyWithValue("dst_cid", "05060708"))
 			})
 
-			It("records connection closes", func() {
-				tracer.ClosedConnection(logging.NewTimeoutCloseReason(logging.TimeoutReasonIdle))
+			It("records the version, if no version negotiation happened", func() {
+				tracer.NegotiatedVersion(0x1337, nil, nil)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("connection_state_updated"))
+				Expect(entry.Name).To(Equal("transport:version_information"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("new", "closed"))
+				Expect(ev).To(HaveLen(1))
+				Expect(ev).To(HaveKeyWithValue("chosen_version", "1337"))
+			})
+
+			It("records the version, if version negotiation happened", func() {
+				tracer.NegotiatedVersion(0x1337, []logging.VersionNumber{1, 2, 3}, []logging.VersionNumber{4, 5, 6})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:version_information"))
+				ev := entry.Event
+				Expect(ev).To(HaveLen(3))
+				Expect(ev).To(HaveKeyWithValue("chosen_version", "1337"))
+				Expect(ev).To(HaveKey("client_versions"))
+				Expect(ev["client_versions"].([]interface{})).To(Equal([]interface{}{"1", "2", "3"}))
+				Expect(ev).To(HaveKey("server_versions"))
+				Expect(ev["server_versions"].([]interface{})).To(Equal([]interface{}{"4", "5", "6"}))
+			})
+
+			It("records idle timeouts", func() {
+				tracer.ClosedConnection(&quic.IdleTimeoutError{})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:connection_closed"))
+				ev := entry.Event
+				Expect(ev).To(HaveLen(2))
+				Expect(ev).To(HaveKeyWithValue("owner", "local"))
 				Expect(ev).To(HaveKeyWithValue("trigger", "idle_timeout"))
 			})
 
-			It("records a received stateless reset packet", func() {
-				tracer.ClosedConnection(logging.NewStatelessResetCloseReason(logging.StatelessResetToken{0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff}))
+			It("records handshake timeouts", func() {
+				tracer.ClosedConnection(&quic.HandshakeTimeoutError{})
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("packet_received"))
+				Expect(entry.Name).To(Equal("transport:connection_closed"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "stateless_reset"))
+				Expect(ev).To(HaveLen(2))
+				Expect(ev).To(HaveKeyWithValue("owner", "local"))
+				Expect(ev).To(HaveKeyWithValue("trigger", "handshake_timeout"))
+			})
+
+			It("records a received stateless reset packet", func() {
+				tracer.ClosedConnection(&quic.StatelessResetError{
+					Token: protocol.StatelessResetToken{0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff},
+				})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:connection_closed"))
+				ev := entry.Event
+				Expect(ev).To(HaveLen(3))
+				Expect(ev).To(HaveKeyWithValue("owner", "remote"))
+				Expect(ev).To(HaveKeyWithValue("trigger", "stateless_reset"))
 				Expect(ev).To(HaveKeyWithValue("stateless_reset_token", "00112233445566778899aabbccddeeff"))
 			})
 
+			It("records connection closing due to version negotiation failure", func() {
+				tracer.ClosedConnection(&quic.VersionNegotiationError{})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:connection_closed"))
+				ev := entry.Event
+				Expect(ev).To(HaveLen(2))
+				Expect(ev).To(HaveKeyWithValue("owner", "remote"))
+				Expect(ev).To(HaveKeyWithValue("trigger", "version_negotiation"))
+			})
+
+			It("records application errors", func() {
+				tracer.ClosedConnection(&quic.ApplicationError{
+					Remote:       true,
+					ErrorCode:    1337,
+					ErrorMessage: "foobar",
+				})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:connection_closed"))
+				ev := entry.Event
+				Expect(ev).To(HaveLen(3))
+				Expect(ev).To(HaveKeyWithValue("owner", "remote"))
+				Expect(ev).To(HaveKeyWithValue("application_code", float64(1337)))
+				Expect(ev).To(HaveKeyWithValue("reason", "foobar"))
+			})
+
+			It("records transport errors", func() {
+				tracer.ClosedConnection(&quic.TransportError{
+					ErrorCode:    qerr.AEADLimitReached,
+					ErrorMessage: "foobar",
+				})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:connection_closed"))
+				ev := entry.Event
+				Expect(ev).To(HaveLen(3))
+				Expect(ev).To(HaveKeyWithValue("owner", "local"))
+				Expect(ev).To(HaveKeyWithValue("connection_code", "aead_limit_reached"))
+				Expect(ev).To(HaveKeyWithValue("reason", "foobar"))
+			})
+
 			It("records sent transport parameters", func() {
 				tracer.SentTransportParameters(&logging.TransportParameters{
 					InitialMaxStreamDataBidiLocal:   1000,
@@ -211,11 +291,11 @@ var _ = Describe("Tracing", func() {
 					InitialSourceConnectionID:       protocol.ConnectionID{0xde, 0xad, 0xbe, 0xef},
 					RetrySourceConnectionID:         &protocol.ConnectionID{0xde, 0xca, 0xfb, 0xad},
 					ActiveConnectionIDLimit:         7,
+					MaxDatagramFrameSize:            protocol.InvalidByteCount,
 				})
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("parameters_set"))
+				Expect(entry.Name).To(Equal("transport:parameters_set"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("owner", "local"))
 				Expect(ev).To(HaveKeyWithValue("original_destination_connection_id", "deadc0de"))
@@ -233,6 +313,7 @@ var _ = Describe("Tracing", func() {
 				Expect(ev).To(HaveKeyWithValue("initial_max_streams_bidi", float64(10)))
 				Expect(ev).To(HaveKeyWithValue("initial_max_streams_uni", float64(20)))
 				Expect(ev).ToNot(HaveKey("preferred_address"))
+				Expect(ev).ToNot(HaveKey("max_datagram_frame_size"))
 			})
 
 			It("records the server's transport parameters, without a stateless reset token", func() {
@@ -242,8 +323,7 @@ var _ = Describe("Tracing", func() {
 				})
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("parameters_set"))
+				Expect(entry.Name).To(Equal("transport:parameters_set"))
 				ev := entry.Event
 				Expect(ev).ToNot(HaveKey("stateless_reset_token"))
 			})
@@ -254,8 +334,7 @@ var _ = Describe("Tracing", func() {
 				})
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("parameters_set"))
+				Expect(entry.Name).To(Equal("transport:parameters_set"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("owner", "local"))
 				Expect(ev).ToNot(HaveKey("retry_source_connection_id"))
@@ -274,8 +353,7 @@ var _ = Describe("Tracing", func() {
 				})
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("parameters_set"))
+				Expect(entry.Name).To(Equal("transport:parameters_set"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("owner", "local"))
 				Expect(ev).To(HaveKey("preferred_address"))
@@ -288,17 +366,51 @@ var _ = Describe("Tracing", func() {
 				Expect(pa).To(HaveKeyWithValue("stateless_reset_token", "0f0e0d0c0b0a09080706050403020100"))
 			})
 
+			It("records transport parameters that enable the datagram extension", func() {
+				tracer.SentTransportParameters(&logging.TransportParameters{
+					MaxDatagramFrameSize: 1337,
+				})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:parameters_set"))
+				ev := entry.Event
+				Expect(ev).To(HaveKeyWithValue("max_datagram_frame_size", float64(1337)))
+			})
+
 			It("records received transport parameters", func() {
 				tracer.ReceivedTransportParameters(&logging.TransportParameters{})
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("parameters_set"))
+				Expect(entry.Name).To(Equal("transport:parameters_set"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("owner", "remote"))
 				Expect(ev).ToNot(HaveKey("original_destination_connection_id"))
 			})
 
+			It("records restored transport parameters", func() {
+				tracer.RestoredTransportParameters(&logging.TransportParameters{
+					InitialMaxStreamDataBidiLocal:  100,
+					InitialMaxStreamDataBidiRemote: 200,
+					InitialMaxStreamDataUni:        300,
+					InitialMaxData:                 400,
+					MaxIdleTimeout:                 123 * time.Millisecond,
+				})
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:parameters_restored"))
+				ev := entry.Event
+				Expect(ev).ToNot(HaveKey("owner"))
+				Expect(ev).ToNot(HaveKey("original_destination_connection_id"))
+				Expect(ev).ToNot(HaveKey("stateless_reset_token"))
+				Expect(ev).ToNot(HaveKey("retry_source_connection_id"))
+				Expect(ev).ToNot(HaveKey("initial_source_connection_id"))
+				Expect(ev).To(HaveKeyWithValue("max_idle_timeout", float64(123)))
+				Expect(ev).To(HaveKeyWithValue("initial_max_data", float64(400)))
+				Expect(ev).To(HaveKeyWithValue("initial_max_stream_data_bidi_local", float64(100)))
+				Expect(ev).To(HaveKeyWithValue("initial_max_stream_data_bidi_remote", float64(200)))
+				Expect(ev).To(HaveKeyWithValue("initial_max_stream_data_uni", float64(300)))
+			})
+
 			It("records a sent packet, without an ACK", func() {
 				tracer.SentPacket(
 					&logging.ExtendedHeader{
@@ -307,6 +419,7 @@ var _ = Describe("Tracing", func() {
 							Type:             protocol.PacketTypeHandshake,
 							DestConnectionID: protocol.ConnectionID{1, 2, 3, 4, 5, 6, 7, 8},
 							SrcConnectionID:  protocol.ConnectionID{4, 3, 2, 1},
+							Length:           1337,
 							Version:          protocol.VersionTLS,
 						},
 						PacketNumber: 1337,
@@ -320,13 +433,15 @@ var _ = Describe("Tracing", func() {
 				)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("packet_sent"))
+				Expect(entry.Name).To(Equal("transport:packet_sent"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "handshake"))
+				Expect(ev).To(HaveKey("raw"))
+				raw := ev["raw"].(map[string]interface{})
+				Expect(raw).To(HaveKeyWithValue("length", float64(987)))
+				Expect(raw).To(HaveKeyWithValue("payload_length", float64(1337)))
 				Expect(ev).To(HaveKey("header"))
 				hdr := ev["header"].(map[string]interface{})
-				Expect(hdr).To(HaveKeyWithValue("packet_size", float64(987)))
+				Expect(hdr).To(HaveKeyWithValue("packet_type", "handshake"))
 				Expect(hdr).To(HaveKeyWithValue("packet_number", float64(1337)))
 				Expect(hdr).To(HaveKeyWithValue("scid", "04030201"))
 				Expect(ev).To(HaveKey("frames"))
@@ -348,8 +463,13 @@ var _ = Describe("Tracing", func() {
 				)
 				entry := exportAndParseSingle()
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "1RTT"))
+				raw := ev["raw"].(map[string]interface{})
+				Expect(raw).To(HaveKeyWithValue("length", float64(123)))
+				Expect(raw).ToNot(HaveKey("payload_length"))
 				Expect(ev).To(HaveKey("header"))
+				hdr := ev["header"].(map[string]interface{})
+				Expect(hdr).To(HaveKeyWithValue("packet_type", "1RTT"))
+				Expect(hdr).To(HaveKeyWithValue("packet_number", float64(1337)))
 				Expect(ev).To(HaveKey("frames"))
 				frames := ev["frames"].([]interface{})
 				Expect(frames).To(HaveLen(2))
@@ -365,6 +485,8 @@ var _ = Describe("Tracing", func() {
 							Type:             protocol.PacketTypeInitial,
 							DestConnectionID: protocol.ConnectionID{1, 2, 3, 4, 5, 6, 7, 8},
 							SrcConnectionID:  protocol.ConnectionID{4, 3, 2, 1},
+							Token:            []byte{0xde, 0xad, 0xbe, 0xef},
+							Length:           1234,
 							Version:          protocol.VersionTLS,
 						},
 						PacketNumber: 1337,
@@ -377,15 +499,20 @@ var _ = Describe("Tracing", func() {
 				)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("packet_received"))
+				Expect(entry.Name).To(Equal("transport:packet_received"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "initial"))
+				Expect(ev).To(HaveKey("raw"))
+				raw := ev["raw"].(map[string]interface{})
+				Expect(raw).To(HaveKeyWithValue("length", float64(789)))
+				Expect(raw).To(HaveKeyWithValue("payload_length", float64(1234)))
 				Expect(ev).To(HaveKey("header"))
 				hdr := ev["header"].(map[string]interface{})
-				Expect(hdr).To(HaveKeyWithValue("packet_size", float64(789)))
+				Expect(hdr).To(HaveKeyWithValue("packet_type", "initial"))
 				Expect(hdr).To(HaveKeyWithValue("packet_number", float64(1337)))
 				Expect(hdr).To(HaveKeyWithValue("scid", "04030201"))
+				Expect(hdr).To(HaveKey("token"))
+				token := hdr["token"].(map[string]interface{})
+				Expect(token).To(HaveKeyWithValue("data", "deadbeef"))
 				Expect(ev).To(HaveKey("frames"))
 				Expect(ev["frames"].([]interface{})).To(HaveLen(2))
 			})
@@ -397,21 +524,25 @@ var _ = Describe("Tracing", func() {
 						Type:             protocol.PacketTypeRetry,
 						DestConnectionID: protocol.ConnectionID{1, 2, 3, 4, 5, 6, 7, 8},
 						SrcConnectionID:  protocol.ConnectionID{4, 3, 2, 1},
+						Token:            []byte{0xde, 0xad, 0xbe, 0xef},
 						Version:          protocol.VersionTLS,
 					},
 				)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("packet_received"))
+				Expect(entry.Name).To(Equal("transport:packet_received"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "retry"))
+				Expect(ev).ToNot(HaveKey("raw"))
 				Expect(ev).To(HaveKey("header"))
-				header := ev["header"]
+				header := ev["header"].(map[string]interface{})
+				Expect(header).To(HaveKeyWithValue("packet_type", "retry"))
 				Expect(header).ToNot(HaveKey("packet_number"))
 				Expect(header).To(HaveKey("version"))
 				Expect(header).To(HaveKey("dcid"))
 				Expect(header).To(HaveKey("scid"))
+				Expect(header).To(HaveKey("token"))
+				token := header["token"].(map[string]interface{})
+				Expect(token).To(HaveKeyWithValue("data", "deadbeef"))
 				Expect(ev).ToNot(HaveKey("frames"))
 			})
 
@@ -427,15 +558,14 @@ var _ = Describe("Tracing", func() {
 				)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("packet_received"))
+				Expect(entry.Name).To(Equal("transport:packet_received"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "version_negotiation"))
 				Expect(ev).To(HaveKey("header"))
 				Expect(ev).ToNot(HaveKey("frames"))
 				Expect(ev).To(HaveKey("supported_versions"))
 				Expect(ev["supported_versions"].([]interface{})).To(Equal([]interface{}{"deadbeef", "decafbad"}))
 				header := ev["header"]
+				Expect(header).To(HaveKeyWithValue("packet_type", "version_negotiation"))
 				Expect(header).ToNot(HaveKey("packet_number"))
 				Expect(header).ToNot(HaveKey("version"))
 				Expect(header).To(HaveKey("dcid"))
@@ -446,10 +576,12 @@ var _ = Describe("Tracing", func() {
 				tracer.BufferedPacket(logging.PacketTypeHandshake)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("packet_buffered"))
+				Expect(entry.Name).To(Equal("transport:packet_buffered"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "handshake"))
+				Expect(ev).To(HaveKey("header"))
+				hdr := ev["header"].(map[string]interface{})
+				Expect(hdr).To(HaveLen(1))
+				Expect(hdr).To(HaveKeyWithValue("packet_type", "handshake"))
 				Expect(ev).To(HaveKeyWithValue("trigger", "keys_unavailable"))
 			})
 
@@ -457,11 +589,14 @@ var _ = Describe("Tracing", func() {
 				tracer.DroppedPacket(logging.PacketTypeHandshake, 1337, logging.PacketDropPayloadDecryptError)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("transport"))
-				Expect(entry.Name).To(Equal("packet_dropped"))
+				Expect(entry.Name).To(Equal("transport:packet_dropped"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "handshake"))
-				Expect(ev).To(HaveKeyWithValue("packet_size", float64(1337)))
+				Expect(ev).To(HaveKey("raw"))
+				Expect(ev["raw"].(map[string]interface{})).To(HaveKeyWithValue("length", float64(1337)))
+				Expect(ev).To(HaveKey("header"))
+				hdr := ev["header"].(map[string]interface{})
+				Expect(hdr).To(HaveLen(1))
+				Expect(hdr).To(HaveKeyWithValue("packet_type", "handshake"))
 				Expect(ev).To(HaveKeyWithValue("trigger", "payload_decrypt_error"))
 			})
 
@@ -485,8 +620,7 @@ var _ = Describe("Tracing", func() {
 				)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("recovery"))
-				Expect(entry.Name).To(Equal("metrics_updated"))
+				Expect(entry.Name).To(Equal("recovery:metrics_updated"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("min_rtt", float64(15)))
 				Expect(ev).To(HaveKeyWithValue("latest_rtt", float64(25)))
@@ -529,12 +663,10 @@ var _ = Describe("Tracing", func() {
 				entries := exportAndParse()
 				Expect(entries).To(HaveLen(2))
 				Expect(entries[0].Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entries[0].Category).To(Equal("recovery"))
-				Expect(entries[0].Name).To(Equal("metrics_updated"))
+				Expect(entries[0].Name).To(Equal("recovery:metrics_updated"))
 				Expect(entries[0].Event).To(HaveLen(7))
 				Expect(entries[1].Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entries[1].Category).To(Equal("recovery"))
-				Expect(entries[1].Name).To(Equal("metrics_updated"))
+				Expect(entries[1].Name).To(Equal("recovery:metrics_updated"))
 				ev := entries[1].Event
 				Expect(ev).ToNot(HaveKey("min_rtt"))
 				Expect(ev).ToNot(HaveKey("congestion_window"))
@@ -547,11 +679,13 @@ var _ = Describe("Tracing", func() {
 				tracer.LostPacket(protocol.EncryptionHandshake, 42, logging.PacketLossReorderingThreshold)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("recovery"))
-				Expect(entry.Name).To(Equal("packet_lost"))
+				Expect(entry.Name).To(Equal("recovery:packet_lost"))
 				ev := entry.Event
-				Expect(ev).To(HaveKeyWithValue("packet_type", "handshake"))
-				Expect(ev).To(HaveKeyWithValue("packet_number", float64(42)))
+				Expect(ev).To(HaveKey("header"))
+				hdr := ev["header"].(map[string]interface{})
+				Expect(hdr).To(HaveLen(2))
+				Expect(hdr).To(HaveKeyWithValue("packet_type", "handshake"))
+				Expect(hdr).To(HaveKeyWithValue("packet_number", float64(42)))
 				Expect(ev).To(HaveKeyWithValue("trigger", "reordering_threshold"))
 			})
 
@@ -559,8 +693,7 @@ var _ = Describe("Tracing", func() {
 				tracer.UpdatedCongestionState(logging.CongestionStateCongestionAvoidance)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("recovery"))
-				Expect(entry.Name).To(Equal("congestion_state_updated"))
+				Expect(entry.Name).To(Equal("recovery:congestion_state_updated"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("new", "congestion_avoidance"))
 			})
@@ -569,8 +702,7 @@ var _ = Describe("Tracing", func() {
 				tracer.UpdatedPTOCount(42)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("recovery"))
-				Expect(entry.Name).To(Equal("metrics_updated"))
+				Expect(entry.Name).To(Equal("recovery:metrics_updated"))
 				Expect(entry.Event).To(HaveKeyWithValue("pto_count", float64(42)))
 			})
 
@@ -578,8 +710,7 @@ var _ = Describe("Tracing", func() {
 				tracer.UpdatedKeyFromTLS(protocol.EncryptionHandshake, protocol.PerspectiveClient)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("security"))
-				Expect(entry.Name).To(Equal("key_updated"))
+				Expect(entry.Name).To(Equal("security:key_updated"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("key_type", "client_handshake_secret"))
 				Expect(ev).To(HaveKeyWithValue("trigger", "tls"))
@@ -592,8 +723,7 @@ var _ = Describe("Tracing", func() {
 				tracer.UpdatedKeyFromTLS(protocol.Encryption1RTT, protocol.PerspectiveServer)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("security"))
-				Expect(entry.Name).To(Equal("key_updated"))
+				Expect(entry.Name).To(Equal("security:key_updated"))
 				ev := entry.Event
 				Expect(ev).To(HaveKeyWithValue("key_type", "server_1rtt_secret"))
 				Expect(ev).To(HaveKeyWithValue("trigger", "tls"))
@@ -609,8 +739,7 @@ var _ = Describe("Tracing", func() {
 				var keyTypes []string
 				for _, entry := range entries {
 					Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-					Expect(entry.Category).To(Equal("security"))
-					Expect(entry.Name).To(Equal("key_updated"))
+					Expect(entry.Name).To(Equal("security:key_updated"))
 					ev := entry.Event
 					Expect(ev).To(HaveKeyWithValue("generation", float64(1337)))
 					Expect(ev).To(HaveKeyWithValue("trigger", "remote_update"))
@@ -628,8 +757,7 @@ var _ = Describe("Tracing", func() {
 				var keyTypes []string
 				for _, entry := range entries {
 					Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-					Expect(entry.Category).To(Equal("security"))
-					Expect(entry.Name).To(Equal("key_retired"))
+					Expect(entry.Name).To(Equal("security:key_retired"))
 					ev := entry.Event
 					Expect(ev).To(HaveKeyWithValue("trigger", "tls"))
 					Expect(ev).To(HaveKey("key_type"))
@@ -639,6 +767,18 @@ var _ = Describe("Tracing", func() {
 				Expect(keyTypes).To(ContainElement("client_initial_secret"))
 			})
 
+			It("records dropped 0-RTT keys", func() {
+				tracer.DroppedEncryptionLevel(protocol.Encryption0RTT)
+				entries := exportAndParse()
+				Expect(entries).To(HaveLen(1))
+				entry := entries[0]
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("security:key_retired"))
+				ev := entry.Event
+				Expect(ev).To(HaveKeyWithValue("trigger", "tls"))
+				Expect(ev).To(HaveKeyWithValue("key_type", "server_0rtt_secret"))
+			})
+
 			It("records dropped keys", func() {
 				tracer.DroppedKey(42)
 				entries := exportAndParse()
@@ -646,8 +786,7 @@ var _ = Describe("Tracing", func() {
 				var keyTypes []string
 				for _, entry := range entries {
 					Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-					Expect(entry.Category).To(Equal("security"))
-					Expect(entry.Name).To(Equal("key_retired"))
+					Expect(entry.Name).To(Equal("security:key_retired"))
 					ev := entry.Event
 					Expect(ev).To(HaveKeyWithValue("generation", float64(42)))
 					Expect(ev).ToNot(HaveKey("trigger"))
@@ -663,8 +802,7 @@ var _ = Describe("Tracing", func() {
 				tracer.SetLossTimer(logging.TimerTypePTO, protocol.EncryptionHandshake, timeout)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("recovery"))
-				Expect(entry.Name).To(Equal("loss_timer_updated"))
+				Expect(entry.Name).To(Equal("recovery:loss_timer_updated"))
 				ev := entry.Event
 				Expect(ev).To(HaveLen(4))
 				Expect(ev).To(HaveKeyWithValue("event_type", "set"))
@@ -672,15 +810,14 @@ var _ = Describe("Tracing", func() {
 				Expect(ev).To(HaveKeyWithValue("packet_number_space", "handshake"))
 				Expect(ev).To(HaveKey("delta"))
 				delta := time.Duration(ev["delta"].(float64)*1e6) * time.Nanosecond
-				Expect(entry.Time.Add(delta)).To(BeTemporally("~", timeout, 10*time.Microsecond))
+				Expect(entry.Time.Add(delta)).To(BeTemporally("~", timeout, scaleDuration(10*time.Microsecond)))
 			})
 
 			It("records when the loss timer expires", func() {
 				tracer.LossTimerExpired(logging.TimerTypeACK, protocol.Encryption1RTT)
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("recovery"))
-				Expect(entry.Name).To(Equal("loss_timer_updated"))
+				Expect(entry.Name).To(Equal("recovery:loss_timer_updated"))
 				ev := entry.Event
 				Expect(ev).To(HaveLen(3))
 				Expect(ev).To(HaveKeyWithValue("event_type", "expired"))
@@ -692,12 +829,21 @@ var _ = Describe("Tracing", func() {
 				tracer.LossTimerCanceled()
 				entry := exportAndParseSingle()
 				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
-				Expect(entry.Category).To(Equal("recovery"))
-				Expect(entry.Name).To(Equal("loss_timer_updated"))
+				Expect(entry.Name).To(Equal("recovery:loss_timer_updated"))
 				ev := entry.Event
 				Expect(ev).To(HaveLen(1))
 				Expect(ev).To(HaveKeyWithValue("event_type", "cancelled"))
 			})
+
+			It("records a generic event", func() {
+				tracer.Debug("foo", "bar")
+				entry := exportAndParseSingle()
+				Expect(entry.Time).To(BeTemporally("~", time.Now(), scaleDuration(10*time.Millisecond)))
+				Expect(entry.Name).To(Equal("transport:foo"))
+				ev := entry.Event
+				Expect(ev).To(HaveLen(1))
+				Expect(ev).To(HaveKeyWithValue("details", "bar"))
+			})
 		})
 	})
 })
